#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import gzip
import json
from typing import cast
from unittest.mock import Mock

import pytest

from pytest_wdl.config import UserConfiguration
from pytest_wdl.core import (
    DefaultDataFile, DataDirs, DataManager, DataResolver, create_data_file
)
from pytest_wdl.localizers import LinkLocalizer, UrlLocalizer
from pytest_wdl.utils import tempdir
from . import GOOD_URL, setenv


def test_data_file():
    with tempdir() as d:
        foo = d / "foo.txt"
        with pytest.raises(ValueError):
            DefaultDataFile(foo, None)

        bar = d / "bar.txt"
        with open(foo, "wt") as out:
            out.write("foo\nbar")
        df = DefaultDataFile(bar, LinkLocalizer(foo))
        assert str(df) == str(bar)

        baz = d / "baz.txt"
        with open(baz, "wt") as out:
            out.write("foo\nbar")
        df.assert_contents_equal(baz)
        df.assert_contents_equal(str(baz))
        df.assert_contents_equal(DefaultDataFile(baz))

        blorf = d / "blorf.txt"
        with open(blorf, "wt") as out:
            out.write("foo\nblorf\nbork")
        with pytest.raises(AssertionError):
            df.assert_contents_equal(blorf)
        df.compare_opts["allowed_diff_lines"] = 1
        with pytest.raises(AssertionError):
            df.assert_contents_equal(blorf)
        df.compare_opts["allowed_diff_lines"] = 2
        df.assert_contents_equal(blorf)


def test_data_file_gz():
    with tempdir() as d:
        foo = d / "foo.txt.gz"
        with gzip.open(foo, "wt") as out:
            out.write("foo\nbar")
        df = DefaultDataFile(foo, allowed_diff_lines=0)

        # Compare identical files
        bar = d / "bar.txt.gz"
        with gzip.open(bar, "wt") as out:
            out.write("foo\nbar")
        df.assert_contents_equal(bar)
        df.assert_contents_equal(str(bar))
        df.assert_contents_equal(DefaultDataFile(bar))

        # Compare differing files
        df.set_compare_opts(allowed_diff_lines=1)
        baz = d / "baz.txt.gz"
        with gzip.open(baz, "wt") as out:
            out.write("foo\nbaz")
        df.assert_contents_equal(bar)
        df.assert_contents_equal(str(bar))
        df.assert_contents_equal(DefaultDataFile(bar))


def test_data_file_dict_type():
    with tempdir() as d:
        foo = d / "foo.txt.gz"
        with gzip.open(foo, "wt") as out:
            out.write("foo\nbar")
        df = create_data_file(
            user_config=UserConfiguration(),
            path=foo,
            type={
                "name": "default",
                "allowed_diff_lines": 1
            }
        )

        bar = d / "bar.txt.gz"
        with gzip.open(bar, "wt") as out:
            out.write("foo\nbaz")

        df.assert_contents_equal(bar)
        df.assert_contents_equal(str(bar))
        df.assert_contents_equal(DefaultDataFile(bar))


def test_data_file_class():
    dd = DataResolver(data_descriptors={
        "foo": {
            "class": "bar",
            "value": 1
        }
    }, user_config=UserConfiguration())
    assert dd.resolve("foo") == 1


def test_data_file_json_contents():
    with tempdir() as d:
        foo = d / "foo.json"
        df = create_data_file(
            user_config=UserConfiguration(),
            path=foo,
            contents={
                "a": 1,
                "b": "foo"
            }
        )
        with open(df.path, "rt") as inp:
            assert json.load(inp) == {
                "a": 1,
                "b": "foo"
            }


def test_data_dirs():
    with tempdir() as d:
        mod = Mock()
        mod.__name__ = "tests.bar"
        cls = Mock()
        cls.__name__ = "baz"
        fun = Mock()
        fun.__name__ = "blorf"
        mod_cls_fun = d / "tests" / "bar" / "baz" / "blorf"
        mod_cls_fun.mkdir(parents=True)
        data_mod_cls_fun = d / "tests" / "data" / "bar" / "baz" / "blorf"
        data_mod_cls_fun.mkdir(parents=True)

        dd = DataDirs(d / "tests", mod, fun, cls)
        assert dd.paths == [
            mod_cls_fun,
            d / "tests" / "bar" / "baz",
            d / "tests" / "bar",
            data_mod_cls_fun,
            d / "tests" / "data" / "bar" / "baz",
            d / "tests" / "data" / "bar",
            d / "tests" / "data"
        ]

        mod_cls_fun = d / "tests" / "bar" / "blorf"
        mod_cls_fun.mkdir(parents=True)
        data_mod_cls_fun = d / "tests" / "data" / "bar" / "blorf"
        data_mod_cls_fun.mkdir(parents=True)
        dd = DataDirs(d / "tests", mod, fun)
        assert dd.paths == [
            mod_cls_fun,
            d / "tests" / "bar",
            data_mod_cls_fun,
            d / "tests" / "data" / "bar",
            d / "tests" / "data"
        ]


def test_data_resolver():
    with tempdir() as d:
        test_data = {
            "foo": {
                "name": "foo.txt"
            },
            "bar": 1
        }
        foo_txt = d / "data" / "foo.txt"
        foo_txt.parent.mkdir()
        with open(foo_txt, "wt") as out:
            out.write("bar")
        mod = Mock()
        mod.__name__ = ""
        fun = Mock()
        fun.__name__ = "test_foo"
        dd = DataDirs(d, mod, fun)
        resolver = DataResolver(test_data, UserConfiguration(None, cache_dir=d))
        with pytest.raises(FileNotFoundError):
            resolver.resolve("bork", dd)
        assert resolver.resolve("foo", dd).path == foo_txt
        assert resolver.resolve("bar", dd) == 1


def test_data_resolver_env():
    with tempdir() as d:
        path = d / "foo.txt"
        with open(path, "wt") as out:
            out.write("foo")
        with setenv({"FOO": str(path)}):
            resolver = DataResolver({
                "foo": {
                    "env": "FOO"
                }
            }, UserConfiguration(None, cache_dir=d))
            assert resolver.resolve("foo").path == path

            bar = d / "bar.txt"
            resolver = DataResolver({
                "foo": {
                    "env": "FOO",
                    "path": bar
                }
            }, UserConfiguration(None, cache_dir=d))
            assert resolver.resolve("foo").path == bar


def test_data_resolver_local_path():
    with tempdir() as d:
        path = d / "foo.txt"
        with open(path, "wt") as out:
            out.write("foo")
        resolver = DataResolver({
            "foo": {
                "path": "foo.txt"
            }
        }, UserConfiguration(None, cache_dir=d))
        assert resolver.resolve("foo").path == path

        with setenv({"MYPATH": str(d)}):
            resolver = DataResolver({
                "foo": {
                    "path": "${MYPATH}/foo.txt"
                }
            }, UserConfiguration(None, cache_dir=d))
            assert resolver.resolve("foo").path == path


def test_data_resolver_create_from_contents():
    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "path": "dir1/dir2/foo.txt",
                "contents": "foo"
            }
        }, UserConfiguration(None, cache_dir=d))
        parent = d / "dir1" / "dir2"
        foo = resolver.resolve("foo")
        assert foo.path == parent / "foo.txt"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"

    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "name": "foo.txt",
                "contents": "foo"
            }
        }, UserConfiguration(None, cache_dir=d))
        foo = resolver.resolve("foo")
        assert foo.path == d / "foo.txt"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"

    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "contents": "foo"
            }
        }, UserConfiguration(None, cache_dir=d))
        foo = resolver.resolve("foo")
        assert foo.path.parent == d
        assert foo.path.exists()
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"


def test_data_resolver_create_from_url():
    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "url": GOOD_URL,
                "path": "dir1/dir2/sample.vcf"
            }
        }, UserConfiguration(None, cache_dir=d))
        foo = resolver.resolve("foo")
        assert foo.path == d / "dir1" / "dir2" / "sample.vcf"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"

    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "url": GOOD_URL,
                "name": "sample.vcf"
            }
        }, UserConfiguration(None, cache_dir=d))
        foo = resolver.resolve("foo")
        assert foo.path == d / "sample.vcf"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"

    with tempdir() as d:
        resolver = DataResolver({
            "foo": {
                "url": GOOD_URL
            }
        }, UserConfiguration(None, cache_dir=d))
        foo = resolver.resolve("foo")
        assert foo.path == d / "test_file"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"


def test_data_resolver_create_from_datadir():
    with tempdir() as d, tempdir() as d1:
        mod = Mock()
        mod.__name__ = "foo.bar"
        cls = Mock()
        cls.__name__ = "baz"
        fun = Mock()
        fun.__name__ = "blorf"
        mod_cls_fun = d / "foo" / "bar" / "baz" / "blorf"
        mod_cls_fun.mkdir(parents=True)
        data_mod_cls_fun = d / "data" / "foo" / "bar" / "baz" / "blorf"
        data_mod_cls_fun.mkdir(parents=True)
        dd = DataDirs(d / "foo", mod, fun, cls)

        resolver = DataResolver({
            "boink": {
                "name": "boink.txt",
            },
            "bobble": {
                "name": "bobble.txt"
            },
            "burp": {
                "name": "burp.txt",
                "path": "burp.txt"
            }
        }, UserConfiguration(None, cache_dir=d1))
        boink = d / "foo" / "bar" / "boink.txt"
        with open(boink, "wt") as out:
            out.write("boink")
        assert boink == resolver.resolve("boink", dd).path

        with pytest.raises(FileNotFoundError):
            resolver.resolve("bobble", dd)

        burp = d / "foo" / "bar" / "burp.txt"
        with open(burp, "wt") as out:
            out.write("burp")
        burp_resolved = resolver.resolve("burp", dd).path
        assert burp_resolved == d1 / "burp.txt"
        assert burp_resolved.is_symlink()

        with pytest.raises(FileNotFoundError):
            resolver.resolve("bobble")


def test_data_manager():
    dm = DataManager(
        data_resolver=DataResolver(
            {
                "foo": {
                    "class": "x",
                    "value": 1
                },
                "bar": {
                    "class": "x",
                    "value": 2
                }
            }, UserConfiguration()
        ),
        datadirs=None
    )
    assert [1, 2] == dm.get_list("foo", "bar")
    assert {"foo": 1, "bork": 2} == dm.get_dict("foo", bork="bar")


def test_http_header_set_in_workflow_data():
    """
    Test that workflow data file can define the HTTP Headers. This is
    important because the URLs referenced can be from different hosts and
    require different headers, so setting them at this level allows that
    fine-grained control.
    """
    with tempdir() as d:
        config = UserConfiguration(cache_dir=d)
        assert not config.default_http_headers
        resolver = DataResolver({
            "foo": {
                "url": GOOD_URL,
                "path": "sample.vcf",
                "http_headers": {
                    "Auth-Header-Token": "TOKEN"
                }
            }
        }, config)
        foo = resolver.resolve("foo")
        assert foo.path == d / "sample.vcf"
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"

    with setenv({"TOKEN": "this_is_the_token"}), tempdir() as d:
        config = UserConfiguration(cache_dir=d)
        assert not config.default_http_headers
        resolver = DataResolver({
            "foo": {
                "url": GOOD_URL,
                "path": "sample.vcf",
                "http_headers": {
                    "Auth-Header-Token": "TOKEN"
                }
            }
        },  config)
        foo = resolver.resolve("foo")
        assert foo.path == d / "sample.vcf"
        assert isinstance(foo.localizer, UrlLocalizer)
        assert cast(UrlLocalizer, foo.localizer).http_headers == {
            "Auth-Header-Token": "this_is_the_token"
        }
        with open(foo.path, "rt") as inp:
            assert inp.read() == "foo"



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
import os
from pathlib import Path
import re
import shutil
import tempfile
from typing import Dict, List, Optional, Union

from pytest_wdl.utils import ensure_path, env_map

try:
    import yaml
except ImportError:
    yaml = None


ENV_USER_CONFIG = "PYTEST_WDL_CONFIG"
DEFAULT_USER_CONFIG_FILE = "pytest_wdl_config"
ENV_CACHE_DIR = "PYTEST_WDL_CACHE_DIR"
KEY_CACHE_DIR = "cache_dir"
ENV_EXECUTION_DIR = "PYTEST_WDL_EXECUTION_DIR"
KEY_EXECUTION_DIR = "execution_dir"
KEY_PROXIES = "proxies"
KEY_HTTP_HEADERS = "http_headers"
KEY_SHOW_PROGRESS = "show_progress"
ENV_DEFAULT_EXECUTORS = "PYTEST_WDL_EXECUTORS"
KEY_DEFAULT_EXECUTORS = "default_executors"
DEFAULT_EXECUTORS = ["miniwdl"]
KEY_EXECUTORS = "executors"


class UserConfiguration:
    """
    Stores pytest-wdl configuration. If configuration options are specified both in
    the config file and as arguments to the constructor, the latter take precedence.

    Args:
        config_file: JSON (or YAML) file from which to load default values.
        cache_dir: The directory in which to cache localized files; defaults to using
            a temporary directory that is specific to each module and deleted
            afterwards.
        remove_cache_dir: Whether to remove the cache directory; if None, takes the
            value True if a temp directory is used for caching, and False, if
            a value for `cache_dir` is specified.
        execution_dir: The directory in which to run workflows. Defaults to None,
            which signals that a different temporary directory should be used for
            each workflow run.
        proxies: Mapping of proxy type (typically 'http' or 'https' to either an
            environment variable, or a dict with either/both keys 'env' and 'value',
            where the value is taken from the environment variable ('env') first, and
            from 'value' if the environment variable is not specified or is unset.
        http_headers: A list of dicts, each of which defines a header. The allowed
            keys are 'pattern', 'name', 'env', and 'value', where pattern is a URL
            pattern to match, 'name' is the header name and 'env' and 'value' are
            interpreted the same as for `proxies`. If no pattern is provided, the
            header is used for all URLs.
        show_progress: Whether to show progress bars when downloading remote test data
            files.
        executors: Default set of executors to run.
        executor_defaults: Mapping of executor name to dict of executor-specific
            configuration options.
    """
    def __init__(
        self,
        config_file: Optional[Path] = None,
        cache_dir: Optional[Path] = None,
        remove_cache_dir: Optional[bool] = None,
        execution_dir: Optional[Path] = None,
        proxies: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
        http_headers: Optional[List[dict]] = None,
        show_progress: Optional[bool] = None,
        executors: Optional[str] = None,
        executor_defaults: Optional[Dict[str, dict]] = None,
        provider_defaults: Optional[Dict[str, dict]] = None,
    ):
        if config_file:
            with open(config_file, "rt") as inp:
                if yaml and config_file.suffix == ".yaml":
                    defaults = yaml.load(inp)
                else:
                    defaults = json.load(inp)
        else:
            defaults = {}

        if not cache_dir:
            cache_dir_str = os.environ.get(ENV_CACHE_DIR, defaults.get(KEY_CACHE_DIR))
            if cache_dir_str:
                cache_dir = ensure_path(cache_dir_str)
        if cache_dir:
            self.cache_dir = ensure_path(cache_dir, is_file=False, create=True)
            if remove_cache_dir is None:
                remove_cache_dir = False
        else:
            self.cache_dir = Path(tempfile.mkdtemp())
            if remove_cache_dir is None:
                remove_cache_dir = True

        self.remove_cache_dir = remove_cache_dir

        if not execution_dir:
            execution_dir_str = os.environ.get(
                ENV_EXECUTION_DIR, defaults.get(KEY_EXECUTION_DIR)
            )
            if execution_dir_str:
                execution_dir = ensure_path(execution_dir_str)

        if execution_dir:
            self.default_execution_dir = ensure_path(
                execution_dir, is_file=False, create=True
            )
        else:
            self.default_execution_dir = None

        if not proxies and KEY_PROXIES in defaults:
            proxies = env_map(defaults[KEY_PROXIES])

        self.proxies = proxies or {}

        if not http_headers and KEY_HTTP_HEADERS in defaults:
            http_headers = defaults[KEY_HTTP_HEADERS]
            for d in http_headers:
                if "pattern" in d:
                    d["pattern"] = re.compile(d.pop("pattern"))

        self.default_http_headers = http_headers or []

        self.show_progress = show_progress
        if self.show_progress is None:
            self.show_progress = defaults.get(KEY_SHOW_PROGRESS)

        if not executors:
            executors_str = os.environ.get(ENV_DEFAULT_EXECUTORS)
            # TODO: test multiple executors specified by environment variable
            if executors_str:
                executors = executors_str.split(",")
            else:
                executors = defaults.get(KEY_DEFAULT_EXECUTORS, DEFAULT_EXECUTORS)

        self.executors = executors

        self.executor_defaults = executor_defaults or {}

        if "executors" in defaults:
            for name, d in defaults["executors"].items():
                name = name.lower()
                if name not in self.executor_defaults:
                    self.executor_defaults[name] = d

        self.provider_defaults = provider_defaults or {}

        if "providers" in defaults:
            for name, d in defaults["providers"].items():
                name = name.lower()
                if name not in self.provider_defaults:
                    self.provider_defaults[name] = d

    def get_executor_defaults(self, executor_name: str) -> dict:
        """
        Get default configuration values for the given executor.

        Args:
            executor_name: The executor name

        Returns:
            A dict with the executor configuration values, if any.
        """
        return self.executor_defaults.get(executor_name.lower(), {})

    def get_provider_defaults(self, provider_name: str) -> dict:
        """
        Get default configuration values for the given provider.

        Args:
            provider_name: The provider name

        Returns:
            A dict with the provider configuration values, if any.
        """
        return self.provider_defaults.get(provider_name.lower(), {})

    def cleanup(self) -> None:
        """
        Preforms cleanup operations, such as deleting the cache directory if
        `self.remove_cache_dir` is True.
        """
        if self.remove_cache_dir:
            shutil.rmtree(self.cache_dir, ignore_errors=True)

    def as_dict(self) -> dict:
        pass  # TODO

    def save(self, path: Path) -> None:
        d = self.as_dict()
        with open(path, "wt") as out:
            if yaml and path.suffix == ".yaml":
                yaml.dump(d, out)
            else:
                json.dump(d, out)


_INSTANCE: Optional[UserConfiguration] = None


def default_user_config_file() -> Path:
    config_file = os.environ.get(ENV_USER_CONFIG)
    config_path = None

    if config_file:
        config_path = ensure_path(config_file)
    else:
        default_config_paths = [
            Path.home() / f"{DEFAULT_USER_CONFIG_FILE}.json",
            Path.home() / f".{DEFAULT_USER_CONFIG_FILE}.json"
        ]
        if yaml:
            default_config_paths.extend([
                Path.home() / f"{DEFAULT_USER_CONFIG_FILE}.yaml",
                Path.home() / f".{DEFAULT_USER_CONFIG_FILE}.yaml"
            ])
        for default_config_path in default_config_paths:
            if default_config_path.exists():
                config_path = default_config_path
                break

    if config_path and not config_path.exists():
        raise FileNotFoundError(f"Config file {config_path} does not exist")

    return config_path


def set_instance(
    config: Optional[UserConfiguration] = None,
    path: Optional[Path] = None,
):
    global _INSTANCE
    if config:
        _INSTANCE = config
    else:
        if not path:
            path = default_user_config_file()
        _INSTANCE = UserConfiguration(path)


def get_instance() -> UserConfiguration:
    global _INSTANCE
    return _INSTANCE


def cleanup():
    global _INSTANCE
    if _INSTANCE:
        _INSTANCE.cleanup()
    _INSTANCE = None



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import os
from pathlib import Path
import zipfile

from pytest_wdl.utils import tempdir
import pytest

from pytest_wdl.utils import ENV_PATH, ENV_CLASSPATH
from pytest_wdl.executors import ENV_JAVA_HOME
from pytest_wdl.executors.cromwell_local import (
    ENV_CROMWELL_CONFIG, ENV_CROMWELL_JAR,  CromwellLocalExecutor
)
from . import setenv, make_executable


def test_java_bin(user_config):
    cromwell_jar_file = user_config.get_executor_defaults("cromwell").get(
        "cromwell_jar_file", os.environ.get(ENV_CROMWELL_JAR)
    )

    with tempdir() as d:
        java = d / "bin" / "java"
        java.parent.mkdir(parents=True)
        with open(java, "wt") as out:
            out.write("foo")
        make_executable(java)

        with setenv({ENV_JAVA_HOME: str(d)}):
            assert  CromwellLocalExecutor(
                [d], cromwell_jar_file=cromwell_jar_file
            ).java_bin == java

        with setenv({
            ENV_PATH: str(d / "bin"),
            ENV_JAVA_HOME: None
        }):
            assert  CromwellLocalExecutor(
                [d], cromwell_jar_file=cromwell_jar_file
            ).java_bin == java

        with setenv({
            ENV_PATH: None,
            ENV_JAVA_HOME: None
        }):
            with pytest.raises(FileNotFoundError):
                assert  CromwellLocalExecutor(
                    [d], cromwell_jar_file=cromwell_jar_file
                ).java_bin

    with setenv({ENV_JAVA_HOME: "foo"}):
        with pytest.raises(FileNotFoundError):
            assert  CromwellLocalExecutor([d]).java_bin


def test_cromwell_config(user_config):
    cromwell_jar_file = user_config.get_executor_defaults("cromwell").get(
        "cromwell_jar_file", os.environ.get(ENV_CROMWELL_JAR)
    )

    with tempdir() as d:
        executor = CromwellLocalExecutor([d], cromwell_jar_file=cromwell_jar_file)
        assert executor._cromwell_args is None
        assert executor.java_args is None

        config = d / "config"
        with setenv({ENV_CROMWELL_CONFIG: str(config)}):
            with pytest.raises(FileNotFoundError):
                 CromwellLocalExecutor([d], cromwell_jar_file=cromwell_jar_file)
            with open(config, "wt") as out:
                out.write("foo")
            executor =  CromwellLocalExecutor(
                [d], cromwell_jar_file=cromwell_jar_file
            )
            assert executor._cromwell_args is None
            assert executor.java_args == f"-Dconfig.file={config}"


def test_java_args(user_config):
    cromwell_jar_file = user_config.get_executor_defaults("cromwell").get(
        "cromwell_jar_file", os.environ.get(ENV_CROMWELL_JAR)
    )

    with tempdir() as d:
        assert  CromwellLocalExecutor(
            [d], cromwell_jar_file=cromwell_jar_file
        ).java_args is None

        with pytest.raises(FileNotFoundError):
             CromwellLocalExecutor(
                [d],
                cromwell_configuration=Path("foo"),
                cromwell_jar_file=cromwell_jar_file
            ).java_args

        config = d / "config"
        with pytest.raises(FileNotFoundError):
             CromwellLocalExecutor(
                [d],
                cromwell_configuration=Path("foo"),
                cromwell_jar_file=cromwell_jar_file,
            ).java_args
        with open(config, "wt") as out:
            out.write("foo")
        assert  CromwellLocalExecutor(
            [d], cromwell_configuration=config, cromwell_jar_file=cromwell_jar_file
        ).java_args == f"-Dconfig.file={config}"


def test_cromwell_jar():
    with tempdir() as d:
        jar = d / "cromwell.jar"

        with setenv({ENV_CROMWELL_JAR: str(jar)}):
            with pytest.raises(FileNotFoundError):
                 CromwellLocalExecutor([d])._cromwell_jar_file
            with open(jar, "wt") as out:
                out.write("foo")
            assert  CromwellLocalExecutor([d])._cromwell_jar_file == jar

        with setenv({
            ENV_CROMWELL_JAR: None,
            ENV_CLASSPATH: str(d)
        }):
            assert  CromwellLocalExecutor([d])._cromwell_jar_file == jar

        with setenv({
            ENV_CROMWELL_JAR: None,
            ENV_CLASSPATH: str(jar)
        }):
            assert  CromwellLocalExecutor([d])._cromwell_jar_file == jar

        with setenv({
            ENV_CROMWELL_JAR: None,
            ENV_CLASSPATH: None
        }):
            with pytest.raises(FileNotFoundError):
                 CromwellLocalExecutor([d])._cromwell_jar_file


def test_get_workflow_imports(user_config):
    cromwell_config = user_config.get_executor_defaults("cromwell")

    with tempdir() as d:
        wdl_dir = d / "foo"
        wdl = wdl_dir / "bar.wdl"
        wdl_dir.mkdir()
        with open(wdl, "wt") as out:
            out.write("foo")
        zip_path =  CromwellLocalExecutor._get_workflow_imports([wdl_dir])
        assert zip_path.exists()
        with zipfile.ZipFile(zip_path, "r") as import_zip:
            names = import_zip.namelist()
            assert len(names) == 1
            assert names[0] == "bar.wdl"
            with import_zip.open("bar.wdl", "r") as inp:
                assert inp.read().decode() == "foo"

    with tempdir() as d:
        wdl_dir = d / "foo"
        wdl = wdl_dir / "bar.wdl"
        wdl_dir.mkdir()
        with open(wdl, "wt") as out:
            out.write("foo")
        imports_file = d / "imports.zip"
        zip_path =  CromwellLocalExecutor._get_workflow_imports(import_dirs=[wdl_dir], imports_file=imports_file)
        assert zip_path.exists()
        assert zip_path == imports_file
        with zipfile.ZipFile(zip_path, "r") as import_zip:
            names = import_zip.namelist()
            assert len(names) == 1
            assert names[0] == "bar.wdl"
            with import_zip.open("bar.wdl", "r") as inp:
                assert inp.read().decode() == "foo"

    with tempdir() as d:
        wdl_dir = d / "foo"
        wdl = wdl_dir / "bar.wdl"
        wdl_dir.mkdir()
        with open(wdl, "wt") as out:
            out.write("foo")
        imports_file = d / "imports.zip"
        with open(imports_file, "wt") as out:
            out.write("foo")
        zip_path =  CromwellLocalExecutor._get_workflow_imports(imports_file=imports_file)
        assert zip_path.exists()
        assert zip_path == imports_file


def test_failure_metadata(workflow_data):
    m44 = workflow_data["metadata44.json"]
    with open(m44.path, "rt") as inp:
        m44_dict = json.load(inp)
    failures =  CromwellLocalExecutor._get_failures(m44_dict)
    assert failures.num_failed == 10
    assert failures.failed_task == "contam_testing_set_org_by_volume.org_blast"
    assert failures.failed_task_exit_status == "Unknown"
    assert "6717f340-b948-4e97-ac2c-bdba793893f5" in \
           str(failures._failed_task_stdout_path)
    assert "6717f340-b948-4e97-ac2c-bdba793893f5" in \
           str(failures._failed_task_stderr_path)

    m47 = workflow_data["metadata47.json"]
    with open(m47.path, "rt") as inp:
        m47_dict = json.load(inp)
    failures =  CromwellLocalExecutor._get_failures(m47_dict)
    assert failures.num_failed == 10
    assert failures.failed_task == "contam_testing_set_org_by_volume.org_blast"
    assert failures.failed_task_exit_status == "Unknown"
    assert "563012ef-e593-42dc-9de0-f0a682ce23e3" in \
           str(failures._failed_task_stdout_path)
    assert "563012ef-e593-42dc-9de0-f0a682ce23e3" in \
           str(failures._failed_task_stderr_path)


def test_call_failure_metadata(workflow_data):
    m = workflow_data["metadata_call_failed.json"]
    with open(m.path, "rt") as inp:
        m_dict = json.load(inp)
    failures =  CromwellLocalExecutor._get_failures(m_dict)
    assert failures.num_failed == 1
    assert failures.failed_task == "ScatterAt27_14"
    assert failures.failed_task_exit_status == "Unknown"
    assert "Failed to evaluate inputs for sub workflow" in failures._failed_task_stderr



def test_variant_caller(workflow_data, workflow_runner):
    inputs = {
        "bam": workflow_data["bam"],
        "ref": {
            "fasta": workflow_data["reference_fa"],
            "organism": "human"
        }
    }
    expected = workflow_data.get_dict("vcf")
    workflow_runner(
        "variant_caller.wdl",
        inputs,
        expected
    )



import json

from pytest_wdl.data_types.json import JsonDataFile
from pytest_wdl.localizers import JsonLocalizer
from pytest_wdl.utils import tempdir


def test_json_data_type():
    with tempdir() as d:
        expected = d / "expected.json"
        actual = d / "actual.json"
        contents = {
            "foo": 1,
            "bar": "a"
        }
        with open(actual, "wt") as out:
            json.dump(contents, out)
        df = JsonDataFile(expected, JsonLocalizer(contents))
        df.assert_contents_equal(actual)



import codecs
import os
from setuptools import setup, find_packages


extras_require = {
    "bam": ["pysam>=0.15.4"],
    "dx": ["dxpy>=0.297.1"],
    "http": ["requests"],
    "progress": ["tqdm"],
    "yaml": ["pyyaml"],
}
extras_require["all"] = list(
    set(lib for lib_array in extras_require.values() for lib in lib_array)
)


setup(
    name="pytest-wdl",
    author="The pytest-wdl development team",
    url="https://github.com/EliLillyCo/pytest-wdl",
    project_urls={
        "Documentation": "https://pytest-wdl.readthedocs.io/en/stable/",
        "Source": "https://github.com/EliLillyCo/pytest-wdl",
        "Tracker": "https://github.com/EliLillyCo/pytest-wdl/issues",
    },
    description="Pytest plugin for testing WDL workflows.",
    long_description_content_type="text/markdown",
    long_description=codecs.open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"),
        "rb",
        "utf-8",
    ).read(),
    license="Apache License 2.0",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    entry_points={
        "pytest11": ["pytest_wdl = pytest_wdl"],
        "pytest_wdl.data_types": [
            "bam = pytest_wdl.data_types.bam:BamDataFile[bam]",
            "vcf = pytest_wdl.data_types.vcf:VcfDataFile",
            "json = pytest_wdl.data_types.json:JsonDataFile",
        ],
        "pytest_wdl.executors": [
            "miniwdl = pytest_wdl.executors.miniwdl:MiniwdlExecutor",
            "cromwell = pytest_wdl.executors.cromwell_local:CromwellLocalExecutor",
            "cromwell-server = pytest_wdl.executors.cromwell_server:"
            "CromwellServerExecutor[http]",
            "dxwdl = pytest_wdl.providers.dx:DxWdlExecutor[dx]",
        ],
        "pytest_wdl.url_schemes": ["dx = pytest_wdl.providers.dx:DxUrlHandler[dx]"],
    },
    py_modules=["pytest_wdl"],
    packages=find_packages(),
    install_requires=[
        "pytest<=5.3.5",
        "subby>=0.1.6",
        "miniwdl==0.7.0",
        "pytest-subtests",
        "xphyle>=4.1.3",
    ],
    extras_require=extras_require,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Pytest",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from abc import ABCMeta, abstractmethod
import json
from pathlib import Path
from typing import Optional, cast
from urllib import request

from xphyle import open_

from pytest_wdl.config import UserConfiguration
from pytest_wdl.url_schemes import Response, ResponseWrapper
from pytest_wdl.utils import (
    LOG, DigestsNotEqualError, env_map, resolve_value_descriptor, verify_digests
)


class Localizer(metaclass=ABCMeta):  # pragma: no-cover
    """
    Abstract base of classes that implement file localization.
    """
    @abstractmethod
    def localize(self, destination: Path) -> None:
        """
        Localize a resource to `destination`.

        Args:
            destination: Path to file where the non-local resource is to be localized.
        """

    def verify(self, path: Path) -> bool:
        """Verify that `path` exists and is valid.

        Args:
            path: Path to verify.

        Returns:
            True if the path is verified, else False
        """
        return path.exists()


class UrlLocalizer(Localizer):
    """
    Localizes a file specified by a URL.
    """
    def __init__(
        self,
        url: str,
        user_config: UserConfiguration,
        http_headers: Optional[dict] = None,
        digests: Optional[dict] = None
    ):
        self.url = url
        self.user_config = user_config
        self._http_headers = http_headers
        self.digests = digests

    def verify(self, path: Path) -> bool:
        if not super().verify(path):
            return False
        if self.digests:
            try:
                verify_digests(path, self.digests)
            except DigestsNotEqualError:
                LOG.exception(
                    "%s already exists but its digest does not match the expected "
                    "digest; deleting the existing file and re-downloading from "
                    "%s", str(path), self.url
                )
                path.unlink()
                return False
        return True

    def localize(self, destination: Path):
        try:
            download_file(
                self.url,
                destination,
                http_headers=self.http_headers,
                proxies=self.user_config.proxies,
                show_progress=self.user_config.show_progress,
                digests=self.digests
            )
        except Exception as err:
            # Delete the destination since it might be incomplete
            if destination.exists():
                try:
                    destination.unlink()
                except IOError:  # TODO: test this
                    LOG.exception(
                        "Error deleting file %s; localization failed, so it may be "
                        "incomplete", str(destination)
                    )
            raise RuntimeError(f"Error localizing url {self.url}") from err

    @property
    def http_headers(self) -> dict:
        http_headers = {}

        if self._http_headers:
            http_headers.update(env_map(self._http_headers))

        if self.user_config.default_http_headers:
            for value_dict in self.user_config.default_http_headers:
                name = value_dict["name"]
                pattern = value_dict.get("pattern")
                if name not in http_headers and (
                    pattern is None or pattern.match(self.url)
                ):
                    value = resolve_value_descriptor(value_dict)
                    if value:
                        http_headers[name] = value

        return http_headers

    @property
    def proxies(self) -> dict:
        return self.user_config.proxies


class StringLocalizer(Localizer):
    """
    Localizes a string by writing it to a file.
    """
    def __init__(self, contents: str):
        self.contents = contents

    def localize(self, destination: Path):
        LOG.debug(f"Persisting {destination} from contents")
        with open_(destination, "wt") as out:
            out.write(self.contents)


class JsonLocalizer(Localizer):
    def __init__(self, contents: dict):
        self.contents = contents

    def localize(self, destination: Path):
        LOG.debug(f"Persisting {destination} from contents")
        with open(destination, "wt") as out:
            json.dump(self.contents, out)


class LinkLocalizer(Localizer):
    """
    Localizes a file to another destination using a symlink.
    """
    def __init__(self, source: Path):
        self.source = source

    def localize(self, destination: Path):
        destination.symlink_to(self.source)


def download_file(
    url: str,
    destination: Path,
    http_headers: Optional[dict] = None,
    proxies: Optional[dict] = None,
    show_progress: bool = True,
    digests: Optional[dict] = None
):
    req = request.Request(url)
    if http_headers:
        for name, value in http_headers.items():
            req.add_unredirected_header(name, value)
    if proxies:
        # TODO: Should we only set the proxy associated with the URL scheme?
        #  Should we raise an exception if there is not a proxy defined for
        #  the URL scheme?
        # parsed = parse.urlparse(url)
        for proxy_type, url in proxies.items():
            req.set_proxy(url, proxy_type)
    rsp = request.urlopen(req)

    if isinstance(rsp, Response):
        downloader = cast(Response, rsp)
    else:
        downloader = ResponseWrapper(rsp)

    LOG.debug("Downloading url %s to %s", url, str(destination))
    downloader.download_file(destination, show_progress, digests)



from collections import defaultdict
import logging
import os
from typing import (
    Dict, Generic, Iterable, Optional, Type, TypeVar, cast
)

from pkg_resources import EntryPoint, ResolutionError, iter_entry_points


LOG = logging.getLogger("pytest-wdl")
LOG.setLevel(os.environ.get("LOGLEVEL", "WARNING").upper())

T = TypeVar("T")


class PluginError(Exception):
    pass


class PluginFactory(Generic[T]):
    """
    Lazily loads a plugin class associated with a data type.
    """
    def __init__(self, entry_point: EntryPoint, return_type: Type[T]):
        self.entry_point = entry_point
        self.return_type = return_type
        self.factory = None

    def __call__(self, *args, **kwargs) -> T:
        if self.factory is None:
            try:
                self.factory = self.entry_point.resolve()
            except ImportError as err:
                raise PluginError(
                    f"Could not load plugin {self.entry_point.name}"
                ) from err

        plugin = self.factory(*args, **kwargs)

        if not isinstance(plugin, self.return_type):  # TODO: test this
            raise RuntimeError(
                f"Expected plugin {plugin} to be an instance of {self.return_type}"
            )

        return cast(self.return_type, plugin)


def plugin_factory_map(
    return_type: Type[T],
    group: Optional[str] = None,
    entry_points: Optional[Iterable[EntryPoint]] = None
) -> Dict[str, PluginFactory[T]]:
    """
    Creates a mapping of entry point name to `PluginFactory` for all discovered
    entry points in the specified group.

    Args:
        group: Entry point group name
        return_type: Expected return type
        entry_points:

    Returns:
        Dict mapping entry point name to `PluginFactory` instances
    """
    if entry_points is None:
        entry_points = iter_entry_points(group=group)

    entry_point_map = defaultdict(list)

    for entry_point in entry_points:
        entry_point_map[entry_point.name].append(entry_point)

    factory_map = {}

    for name, points in entry_point_map.items():
        if len(points) > 1:
            # Filter out built-ins
            points = list(filter(
                lambda point: not point.module_name.startswith("pytest_wdl"), points
            ))
            if len(points) > 1:
                raise RuntimeError(
                    f"Multiple third-party plugins found in group {group} with the "
                    f"same name: {name}"
                )

        ep = points[0]

        try:
            ep.require()
        except ResourceWarning as rerr:
            LOG.warning(
                "Plugin %s is not available because it is missing an extra "
                "dependency: %s", name, str(rerr)
            )
            continue
        except ResolutionError as rerr:
            LOG.warning(
                "Plugin %s is not available because it is missing an extra "
                "dependency: %s", name, str(rerr)
            )
            continue
        except PluginError as perr:
            LOG.warning(
                "Error while loading plugin %s: %s", name, str(perr)
            )
            continue

        factory_map[name] = PluginFactory(ep, return_type)

    return factory_map
